(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import sepalg.
Require Import sepalg_generators.
Require Import cjoins.

Definition cross_split `{sepalg A} :=
      forall a b c d z : A,
       join a b z ->
       join c d z ->
    { x:(A*A*A*A) &  match x with (ac,ad,bc,bd) =>
         join ac ad a /\
         join bc bd b /\
         join ac bc c /\
         join ad bd d
       end
    }.

(** The cross split axiom looks unwieldly,
    but here we show that it arises naturally
    as a kind of distributivity property.
    Cross split can be rendered, with some accuracy,
    as "the separation algebra is distributive."
 *)

  (** This definition mirrors the definition of
      distributivity in a join-semilattice.  This
      definition generalizes the standard notion of
      distributivity in a lattice, but only mentions
      of the lattice operators.  Here we transplant
      the semilattice definition into the setting
      of separation algebras.
    *)
  Definition sa_distributive `{sepalg A} :=
    forall a b x z,
      join a b z ->
      constructive_join_sub x z ->
      {a' : A & {b' : A & 
           (constructive_join_sub a' a * constructive_join_sub b' b * join a' b' x)%type}}.

(*
  (** We define this weaker version of cross-split
      in order to show that the sa_distributive
      axiom is equivelant. The ordinary cross_split
      is more constructive (it uses a sigma type rather
      than 'exists'), so we have to weaken it to show
      the correspondence.  We could, instead, define
      and use a constructive version of join_sub.
    *)
  Definition weak_cross_split `{sepalg A} :=
    forall a b c d z : A,
      join a b z ->
      join c d z ->
      exists x:(A*A*A*A), match x with (ac,ad,bc,bd) =>
         join ac ad a /\
         join bc bd b /\
         join ac bc c /\
         join ad bd d
       end.
*)

  (** Here we show that the cross split axiom is
      the same as the statement of distributivity
      for join semilattices transliterated into the
      setting of separation algebras.
    *)
  Theorem cross_split_distibutive `{sepalg A} :
     cross_split -> sa_distributive.
  Proof.
    unfold sa_distributive; intros A H H0 ? ? ? ? H1 [x0 H2].
    destruct (H0 _ _ _ _ _ H1 H2).
    destruct x1.
    destruct p.
    destruct p.
    intuition eauto.
    exists a2.
    exists a1.
    intuition eauto.
    econstructor; eauto.
    econstructor; eauto.
  Qed.

  Theorem distributive_cross_split `{sepalg A}:
     sa_distributive -> cross_split.
  Proof.
    intros A H H0.
    repeat intro.
    hnf in H0.
    destruct (H0 a b c z H1) as [a' [b' [[?H ?H] ?H]]].
    exists d; auto.
    destruct H3 as [q ?H].
    destruct H4 as [w ?H].
    exists (a',q,b',w); intuition.
    destruct (join_assoc H3 H1) as [f [? ?]].
    apply join_com in H6.
    destruct (join_assoc H4 H6) as [g [? ?]].
    replace d with g.
    apply join_com; auto.
    apply join_com in H7.
    apply join_com in H9.
    destruct (join_assoc H9 H7) as [h [? ?]].
    assert (c = h).
    eapply join_eq; eauto.
    subst h.
    eapply join_canc; eauto.
  Qed.

(** NOTICE ABOUT REDUNDANT LEMMAS:
 Since sa_distribute <-> cross_split, many of the proofs below are redundant.
 This was part of an experiment to see whether, in general, sa_distributive is
 simpler to prove than cross_split.  Short answer:  not really.
*)

Lemma distributive_equiv: forall A, @sa_distributive A (sa_equiv A).
Proof.
  repeat intro.
 destruct H; subst.
 exists x; exists x; repeat split; auto.  
Qed.

Lemma cross_split_equiv : forall A,
  @cross_split A (sa_equiv A).
Proof. intros; apply distributive_cross_split; apply distributive_equiv. Qed.

Lemma distributive_fun: forall A (sa: sepalg A) (key: Type),
               @sa_distributive _ sa -> @sa_distributive _ (sa_fun key sa).
Proof.
unfold sa_distributive; intros.
assert (forall k, constructive_join_sub (x k) (z k)).
destruct X0 as [y ?].
intro k; exists (y k); auto.
assert (J := fun (k: key) => X (a k) (b k) (x k) (z k) (H k) (X1 k)).
clear X.
exists (fun k => projT1 (J k)).
exists (fun k => projT1 (projT2 (J k))).
split; [split|].
exists (fun k => projT1 (fst (fst (projT2 (projT2 (J k))))));
intro k; destruct (J k) as [ak' [bk' [[c c0] j]]];  simpl; destruct c; auto.
exists (fun k => projT1 (snd (fst (projT2 (projT2 (J k))))));
intro k; destruct (J k) as [ak' [bk' [[c c0] j]]];  simpl; destruct c0; auto.
intro k; destruct (J k) as [ak' [bk' [[c c0] j]]]; simpl; auto.
Qed.

Lemma cross_split_fun: forall A (sa: sepalg A) (key: Type),
          @cross_split _ sa -> @cross_split _ (sa_fun key sa).
Proof.
intros.
unfold cross_split in *.
intros.
pose (f (x: key) := projT1 (X (a x) (b x) (c x) (d x) (z x) (H x) (H0 x))).
pose (g (x: key) := projT2 (X (a x) (b x) (c x) (d x) (z x) (H x) (H0 x))).
pose (ac (x: key) := fst (fst (fst (f x)))).
pose (ad (x: key) := snd (fst (fst (f x)))).
pose (bc (x: key) := snd (fst (f x))).
pose (bd (x: key) := snd (f x)).
exists (ac,ad,bc,bd).
unfold ac, ad, bc, bd, f; clear ac ad bc bd f.
repeat split; intro x; simpl;
generalize (g x);  destruct (projT1 (X (a x) (b x) (c x) (d x) (z x) (H x) (H0 x))) as [[[? ?] ?] ?]; simpl; intuition.
Qed.

Lemma sa_distributive_prod : forall A B saA saB,
  @sa_distributive A saA ->
  @sa_distributive B saB ->
  @sa_distributive (A * B) (sa_prod saA saB).
Proof.
intros.
 intros [a1 a2] [b1 b2] [c1 c2] [z1 z2] [? ?].
 intros [[d1 d2] [? ?]].
 simpl in *.
 destruct (X a1 b1 c1 z1 H (existT _ d1 H1)) as [a1' [b1' [[[u1 ?] [v1 ?]] ?]]].
 destruct (X0 a2 b2 c2 z2 H0 (existT _ d2 H2)) as [a2' [b2' [[[u2 ?] [v2 ?]] ?]]].
 exists (a1',a2'). exists (b1',b2').
 split; [split|].
 exists (u1,u2); split; auto.
 exists (v1,v2); split; auto.
 split; auto.
Qed.

Lemma cross_split_prod : forall A B saA saB,
  @cross_split A saA ->
  @cross_split B saB ->
  @cross_split (A * B) (sa_prod saA saB).
Proof.
  repeat intro.
  destruct a as [a1 a2].
  destruct b as [b1 b2].
  destruct c as [c1 c2].
  destruct d as [d1 d2].
  destruct z as [z1 z2].
  destruct H.
  destruct H0.
  simpl in *.
  destruct (X a1 b1 c1 d1 z1)
    as [p ?]; auto.
  destruct p as [[[s1 p1] q1] r1].
  destruct (X0 a2 b2 c2 d2 z2)
    as [p ?]; auto.
  destruct p as [[[s2 p2] q2] r2].
  exists ((s1,s2),(p1,p2),(q1,q2),(r1,r2)).
  unfold join_prod; simpl; intuition.
Qed.

Lemma sa_distributive_bij : forall A B saA bij,
  @sa_distributive A saA ->
  @sa_distributive B (sa_bijection A B bij saA).
Proof.
 repeat intro.
 destruct X0 as [u ?].
 destruct bij.
 destruct (X (bij_g a) (bij_g b) (bij_g x) (bij_g z)) as [a' [b' [[[? ?] [? ?]] ?]]]; auto.
 exists (bij_g u); auto.
 exists (bij_f a'); exists (bij_f b'); split; [split|].
 exists (bij_f x0); simpl; repeat rewrite bij_gf; auto.
 exists (bij_f x1); simpl; repeat rewrite bij_gf; auto.
 simpl; repeat rewrite bij_gf; auto.
Qed. 

Lemma cross_split_bij : forall A B saA bij,
  @cross_split A saA ->
  @cross_split B (sa_bijection A B bij saA).
Proof.
  repeat intro.
  hnf in H, H0.
  destruct bij.
  destruct (X (bij_g a) (bij_g b) (bij_g c) (bij_g d) (bij_g z)); auto.
  destruct x as [[[s p] q] r].
  exists (bij_f s,bij_f p,bij_f q,bij_f r).
  simpl.
  repeat rewrite bij_gf.
  auto.
Qed.

Lemma constructive_join_sub_pjoinable `{sa: sepalg A}:
  (forall x:A, {identity x}+{~identity x}) ->
  forall a c : pjoinable sa,
    constructive_join_sub (projT1 a) (projT1 c) ->
    @constructive_join_sub _ (sa_lift sa) (Some a) (Some c).
Proof.
intros.
destruct X0 as [b ?].
destruct (X b).
assert (a=c).
destruct a; destruct c. apply exist_ext.
simpl in j.
eapply join_eq; try apply j. apply join_com; apply identity_unit; eauto.
subst c.
exists None; constructor.
exists (Some (exist nonidentity _ n)).
constructor.
destruct a; destruct c; simpl in *.
auto.
Qed.

Lemma sa_distributive_lift : forall A saA,
  (forall x:A, {identity x}+{~identity x}) ->
  @sa_distributive A saA ->
  @sa_distributive _ (sa_lift saA).
Proof.
intros.
intros [[a Ha]|].
2: intros; assert (b=z) by (inv H; auto); subst z; exists None; exists x;
    split; [split|]; auto; constructor.
intros [[b Hb]|].
2: intros b [[z Hz]|] ? ?; 
 [assert (a=z) by (inv H; auto); subst z; clear H;
   rewrite (proof_irr Hz Ha) in X1; clear Hz; exists b; exists None;
    split; [split|]; auto; constructor
 | elimtype False; inv H].
intros [[c Hc]|].
2: intros ? ? ?; exists None; exists None; split; [split|]; econstructor; econstructor.
intros [[z Hz]|] H Hj.
2: elimtype False; inv H.
destruct (X0 a b c z) as [a' [b' [[? ?] ?]]].
inversion H; auto.
inversion Hj; auto.
destruct x as [[x Hx]|].
exists x; inv H0; auto.
replace z with c; auto.
inv H0; auto.
destruct (X a') as [Pa'|Pa']; [exists None | exists (Some (exist nonidentity _ Pa'))].
assert (b'=c) by (eapply join_eq; try apply j; apply identity_unit; eauto).
subst b'.
exists (Some (exist nonidentity c Hc)).
split; [split|]; eauto. econstructor; econstructor.
apply constructive_join_sub_pjoinable; auto.
constructor.
destruct (X b') as [Pb'|Pb']; [exists None | exists (Some (exist nonidentity _ Pb'))].
split; [split|]; eauto.
apply constructive_join_sub_pjoinable; auto.
econstructor; econstructor.
apply join_unit2. econstructor; eauto.
f_equal. apply exist_ext.
symmetry. eapply join_eq. eapply join_com; apply j.  apply identity_unit; eauto.
split; [split|]; eauto.
apply constructive_join_sub_pjoinable; auto.
apply constructive_join_sub_pjoinable; auto.
constructor. simpl. auto.
Qed.

Lemma cross_split_lift : forall A saA,
  (forall x:A, {identity x}+{~identity x}) ->
  @cross_split A saA ->
  @cross_split _ (sa_lift saA).
Proof.
  intros.
  hnf; intros.
  destruct z as [[z Hz]|].
  destruct a as [[a Ha]|].
  destruct b as [[b Hb]|].
  simpl in *.
  assert (join a b z).
  inv H; auto.
  destruct c as [[c Hc]|].
  destruct d as [[d Hd]|].
  assert (join c d z).
  inv H0; auto.
  destruct (X0 a b c d z H1 H2).
  destruct x as [[[s p] q] r].
  set (f w :=
       match X w with
       | left _ => None
       | right H => Some (exist (fun x => ~identity x) w H)
       end).
  exists (f s, f p, f q, f r).
  unfold f.
  split; simpl.
  destruct (X s); destruct (X p).
  elim Ha.
  destruct y.
  replace a with p; auto.
  assert (p = a).
  destruct y.
  apply i; auto.
  subst a.
  replace Ha with n.
  constructor.
  extensionality x.
  intros; elim (n x).
  assert (s = a).
  apply i.
  apply join_com; intuition.
  subst a.
  replace Ha with n.
  constructor.
  extensionality x.
  intros; elim (n x).
  constructor.
  simpl; intuition.
  split.
  destruct (X q); destruct (X r).
  elim Hb.
  replace b with r; auto.
  apply i; intuition.
  assert (r = b).
  apply i; intuition.
  subst b.
  replace Hb with n.
  constructor.
  extensionality x.
  intros; elim (n x).
  assert (q = b).
  apply i; intuition.
  subst b.
  replace Hb with n.
  constructor.
  extensionality x.
  intros; elim (n x).
  constructor; simpl; intuition.
  split.
  destruct (X s); destruct (X q).
  assert (q = c).
  apply i; intuition.
  elim Hc.
  subst c; auto.
  assert (q = c).
  apply i; intuition.
  subst c.
  replace Hc with n.
  constructor.
  extensionality x.
  intros; elim (n x).
  assert (s = c).
  apply i; intuition.
  subst c.
  replace Hc with n.
  constructor.
  extensionality x.
  intros; elim (n x).
  constructor; simpl; intuition.
  destruct (X p); destruct (X r).
  elim Hd.
  replace d with r; auto.
  apply i; intuition.
  assert (r = d).
  apply i; intuition.
  subst d.
  replace Hd with n.
  constructor.
  extensionality x.
  intros; elim (n x).
  assert (p = d).
  apply i; intuition.
  subst d.
  replace Hd with n.
  constructor.
  extensionality x.
  intros; elim (n x).
  constructor; simpl; intuition.

  exists (Some (exist _ a Ha),None,Some (exist _ b Hb),None).
  intuition; constructor.
  simpl.
  inv H0; auto.
  exists (None,Some (exist _ a Ha),None,Some (exist _ b Hb)).
  intuition; try constructor.
  inv H0.
  constructor.
  simpl; auto.

  assert (a = z).
  inv H; auto.
  subst a.
  exists (c,d,None,None).
  intuition; try constructor.
  replace Ha with Hz; auto.
  extensionality x.
  intros; elim (Hz x).
  exists (None,None,c,d).
  intuition; try constructor.
  inv H; auto.
  
  exists (None,None,None,None).
  inv H; inv H0; intuition constructor.
Qed.


Lemma cross_split_fpm : forall A Adec B saB,
  (forall x:B, {identity x}+{~identity x}) ->
  @cross_split B saB ->
  @cross_split _ (fpm_sa A Adec B saB).
Proof.
  intros.
  
  hnf; intros.
  destruct a as [a Ha].
  destruct b as [b Hb].
  destruct c as [c Hc].
  destruct d as [d Hd].
  destruct z as [z Hz].
  hnf in H, H0; simpl in *.

  assert (@cross_split (A -> option (pjoinable saB)) (sa_fun A (sa_lift saB))).
  apply cross_split_fun.
  apply cross_split_lift; auto.
  destruct (X1 a b c d z); auto.
  destruct x as [[[s p] q] r].
  decompose [and] y; clear y.
  assert (Hs : finMap _ _ _ s).
  destruct Ha.
  exists x.
  intros.
  hnf in H1.
  spec H1 a0.
  rewrite H4 in H1; auto.
  inv H1; auto.
  assert (Hp: finMap _ _ _ p).
  destruct Ha.
  exists x.
  intros.
  spec H1 a0.
  rewrite H4 in H1; auto.
  inv H1; auto.
  assert (Hq : finMap _ _ _ q).
  destruct Hb.
  exists x.
  intros.
  spec H3 a0.
  rewrite H4 in H3; auto.
  inv H3; auto.
  assert (Hr : finMap _ _ _ r).
  destruct Hb.
  exists x.
  intros.
  spec H3 a0.
  rewrite H4 in H3; auto.
  inv H3; auto.
  exists (exist _ s Hs, exist _ p Hp, exist _ q Hq, exist _ r Hr).
  simpl; intuition.
Qed.
