(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import sepalg.
Require Import sepalg_generators.
Require Import sepalg_functors.
Require Import ageable.
Require Import age_sepalg.
Require Import knot.
Require Import knot_lemmas.

Module Type TY_FUNCTOR_SA.
  Declare Module TF:TY_FUNCTOR.
  Import TF.
 
  Parameter F_sa : forall A, sepalg A -> sepalg (F A).
  Implicit Arguments F_sa.

  Axiom fmap_hom : forall A B sa1 sa2 (f:A->B),
    @join_hom _ sa1 _ sa2 f ->
    @join_hom _ (F_sa sa1) _ (F_sa sa2) (fmap f).

  Instance T_sa : sepalg T.
  Axiom T_bot_identity : identity T_bot.

  Axiom F_preserves_unmaps : preserves_unmaps F F_sa fmap.

  Axiom F_identities : forall A B (saA:sepalg A) (saB:sepalg B) (f:A->B) x,
    @identity _ (F_sa saB) (fmap f x) -> 
    @identity _ (F_sa saA) x.

End TY_FUNCTOR_SA.

Module Type KNOT_SA.
  Declare Module TFSA:TY_FUNCTOR_SA.
  Declare Module K:KNOT with Module TF:=TFSA.TF.

  Import TFSA.TF.
  Import TFSA.
  Import K.

  Parameter knot_sa : sepalg knot.
  Existing Instance knot_sa.

  Definition nat_F_sa : sepalg (nat * F predicate) :=
    sa_prod (sa_equiv nat) (F_sa (sa_fun (knot * other) T_sa)).
  Existing Instance nat_F_sa.

  Axiom join_unsquash : forall x1 x2 x3,
    join x1 x2 x3 =
    join (unsquash x1) (unsquash x2) (unsquash x3).

  Axiom knot_asa : @ASA knot knot_sa K.ageable_knot.

End KNOT_SA.

Module KnotSa (TFSA':TY_FUNCTOR_SA) (K':KNOT with Module TF:=TFSA'.TF)
  : KNOT_SA with Module TFSA:=TFSA' with Module K:=K'.

  Module TFSA:=TFSA'.
  Module K:=K'.
  Module KL := Knot_Lemmas(K).

  Import TFSA.TF.
  Import TFSA.
  Import K.
  Import KL.

  Definition pred_sa : sepalg predicate := sa_fun _ T_sa.
  Existing Instance pred_sa.

  Lemma approx_join_hom : forall n, 
    join_hom (approx n).
  Proof.
    unfold join_hom.
    intros.
    hnf in *.
    intro x0.
    spec H x0.
    unfold K'.approx.
    destruct (le_gt_dec n (level (fst x0))).
    generalize (T_bot_identity); intro.
    rewrite identity_unit_equiv in H0.
    apply H0.
    trivial.
  Qed.

  Lemma F_approx_join_hom : forall n,
    join_hom (H := F_sa pred_sa) (H0 := F_sa pred_sa) (fmap (approx n)).
  Proof.
    intros; apply fmap_hom; apply approx_join_hom.
  Qed.
  
  Definition nat_F_sa : sepalg (nat * F predicate) :=
    sa_prod (sa_equiv nat) (F_sa pred_sa).
  Existing Instance nat_F_sa.

  Lemma unsquash_squash_join_hom : join_hom (unsquash oo squash).
  Proof.
    unfold compose.
    intros [x1 x2] [y1 y2] [z1 z2] ?.
    do 3 rewrite (unsquash_squash).
    firstorder.
    simpl in *.
    subst y1.
    subst z1.
    apply F_approx_join_hom.
    trivial.
  Qed.

  Definition knot_sa := 
    preimage_sa _ _ nat_F_sa unsquash squash squash_unsquash unsquash_squash_join_hom.
  Existing Instance knot_sa.

  Lemma join_unsquash : forall x1 x2 x3,
    join x1 x2 x3 =
    join (unsquash x1) (unsquash x2) (unsquash x3).
  Proof.
    intuition.
  Qed.

  Lemma age_join1 :
    forall x y z x' : K'.knot,
      join x y z ->
      age x x' ->
      exists y' : K'.knot,
        exists z' : K'.knot, join x' y' z' /\ age y y' /\ age z z'.
  Proof.
    intros.
    unfold age in *; simpl in *.
    rewrite knot_age1 in H0.
    repeat rewrite knot_age1.
    destruct (unsquash x).
    destruct (unsquash y).
    destruct (unsquash z).
    simpl in H; destruct H.
    simpl in H; destruct H.
    subst n0 n1.
    destruct n; try discriminate.
    exists (squash (n,f0)).
    exists (squash (n,f1)).
    intuition.
    inv H0.
    simpl in *.
    repeat rewrite unsquash_squash.
    split; simpl.
    split; simpl; auto.
    apply F_approx_join_hom; auto.
  Qed.

  Lemma age_join2 :
    forall x y z z' : K'.knot,
      join x y z ->
      age z z' ->
      exists x' : K'.knot,
        exists y' : K'.knot, join x' y' z' /\ age x x' /\ age y y'.
  Proof.
    intros.
    unfold age in *; simpl in *.
    rewrite knot_age1 in H0.
    repeat rewrite knot_age1.
    destruct (unsquash x).
    destruct (unsquash y).
    destruct (unsquash z).
    destruct n1; try discriminate.
    inv H0.
    destruct H; simpl in *.
    destruct H; subst.
    exists (squash (n1,f)).
    exists (squash (n1,f0)).
    intuition.
    repeat rewrite unsquash_squash.
    split; simpl.
    split; simpl; auto.
    apply F_approx_join_hom; auto.
  Qed.

  Lemma pred_unmap_left_approx :
    forall n, unmap_left pred_sa pred_sa (approx n).
  Proof.
    red; intros.
    assert (x' = approx n x').
      extensionality k.
      spec H k.
      unfold approx in *.
      revert H.
      elim (le_gt_dec n (level (fst k))); intros; trivial.
      apply T_bot_identity.
      apply join_com.
      trivial.
    exists (fun w => if le_gt_dec n (level (fst w)) then projT1 (join_ex_identities (z w)) else x' w).
    exists (fun w => if le_gt_dec n (level (fst w)) then z w else y w).
    split.
    intro w.
    spec H w.
    revert H.
    rewrite H0.
    unfold approx.
    elim (le_gt_dec n (level (fst w))); intros; trivial.
    destruct (join_ex_identities (z w)).
    simpl.
    destruct a0 as [? [? ?]].
    generalize (H1 _ _ H2); intro.
    subst x0.
    trivial.
    
    split.
    extensionality w.
    spec H w.
    revert H.
    unfold approx.
    elim (le_gt_dec n (level (fst w))); intros; trivial.
    apply (@join_canc _ _ T_bot (x' w) T_bot T_bot); trivial.
    generalize T_bot_identity; intro.
    rewrite identity_unit_equiv in H1.
    apply H1.
    
    unfold approx.
    extensionality w.
    elim (le_gt_dec n (level (fst w))); intros; trivial.
  Qed.

  Lemma pred_unmap_right_approx :
    forall n, unmap_right pred_sa pred_sa (approx n).
  Proof.
    red; intros.
    exists (fun w => if le_gt_dec n (level (fst w)) then (projT1 (join_ex_identities (x w))) else y w).
    exists (fun w => if le_gt_dec n (level (fst w)) then x w else z' w).
    split.
    intro w.
    spec H w.
    revert H.
    unfold approx.
    elim (le_gt_dec n (level (fst w))); intros; trivial.
    destruct (join_ex_identities (x w)).
    simpl.
    destruct a0 as [? [? ?]].
    generalize (H0 _ _ H1); intro.
    subst x1.
    apply join_com.
    trivial.
    
    split.
    unfold approx.
    extensionality w.
    elim (le_gt_dec n (level (fst w))); intros; trivial.
    
    pattern z' at 2.
    assert (z' = approx n z').
      extensionality k.
      spec H k.
      unfold approx in *.
      revert H.
      elim (le_gt_dec n (level (fst k))); intros; trivial.
      symmetry.
      apply (T_bot_identity _ _ H).
    rewrite H0.
    unfold approx.
    extensionality w.
    elim (le_gt_dec n (level (fst w))); intros; trivial.
  Qed.

  Lemma unage_join1 : forall x x' y' z', join x' y' z' -> age x x' ->
    exists y, exists z, join x y z /\ age y y' /\ age z z'.
  Proof.
    destruct F_preserves_unmaps as [_ fmap_commute1].
    intros.
    unfold age in *; simpl in *.
    rewrite knot_age1 in H0.
    case_eq (unsquash x); intros.
    rewrite H1 in H0.
    destruct n; try discriminate.
    inv H0.
    revert H.
    case_eq (unsquash y');
    case_eq (unsquash z'); intros.
    rewrite unsquash_squash in H2.
    destruct H2; simpl in *.
    destruct H2; subst.
    rename n0 into n.
    destruct (fmap_commute1 _ _ pred_sa pred_sa (approx n) (approx_join_hom n) (pred_unmap_right_approx _) f f1 f0) as [q [w [? [? ?]]]]; auto.
    rewrite <- (unsquash_approx H0); auto.
    exists (squash (S n,q)).
    exists (squash (S n,w)).
    repeat rewrite unsquash_squash.

    split.
    split; simpl; auto.
    split; simpl; auto.
    generalize (F_approx_join_hom (S n) _ _ _ H2).
    rewrite <- (unsquash_approx H1); auto.

    split.
    rewrite knot_age1.
    rewrite unsquash_squash.
    replace y' with (squash (n,fmap (approx (S n)) q)); auto.
    apply unsquash_inj.
    rewrite unsquash_squash, H0.
    apply injective_projections; simpl; auto.
    rewrite (unsquash_approx H0).
    rewrite <- H4.
    change ((fmap (approx n) oo fmap (approx (S n))) q = fmap (approx n) q).
    rewrite fmap_comp.
    replace (approx n oo approx (S n)) with (approx n); auto.
    extensionality a.
    replace (S n) with (1 + n)%nat by trivial.
    rewrite <- (approx_approx1 1 n).
    trivial.

    rewrite knot_age1.
    rewrite unsquash_squash.
    replace z' with  (squash (n,fmap (approx (S n)) w)); auto.
    apply unsquash_inj.
    rewrite unsquash_squash, H.
    apply injective_projections; simpl; auto.
    rewrite <- H5.
    change ((fmap (approx n) oo fmap (approx (S n))) w = fmap (approx n) w).
    rewrite fmap_comp.
    replace (approx n oo approx (S n)) with (approx n); auto.
    extensionality a.
    replace (S n) with (1 + n)%nat by trivial.
    rewrite <- (approx_approx1 1 n).
    trivial.
  Qed.

  Lemma unage_join2 :
    forall z x' y' z', join x' y' z' -> age z z' ->
      exists x, exists y, join x y z /\ age x x' /\ age y y'.
  Proof.
    destruct F_preserves_unmaps as [fmap_commute2 fmap_commute1].
    intros.
    rewrite join_unsquash in H.
    revert H H0.
    unfold age in *; simpl in *.
    repeat rewrite knot_age1.
    case_eq (unsquash x');
    case_eq (unsquash y');
    case_eq (unsquash z');
    case_eq (unsquash z); intros.
    destruct n; try discriminate.
    inv H4.
    rewrite unsquash_squash in H0.
    inv H0.
    destruct H3; simpl in *.
    destruct H0; subst.
    rename n0 into n.
    destruct (fmap_commute2 _ _ pred_sa pred_sa (approx n) (approx_join_hom n) (pred_unmap_left_approx _) f2 f1 f)
      as [wx [wy [? [? ?]]]]; auto.
    rewrite <- (unsquash_approx H1); auto.
    exists (squash (S n, wx)).
    exists (squash (S n, wy)).
    repeat rewrite unsquash_squash.

    split.
    split; simpl; auto.
    split; simpl; auto.
    rewrite (unsquash_approx H).
    apply F_approx_join_hom; auto.
    split; rewrite knot_age1; rewrite unsquash_squash.
    replace x' with (squash (n,(fmap (approx (S n)) wx))); auto.
    apply unsquash_inj.
    rewrite unsquash_squash, H2.
    apply injective_projections; simpl; auto.
    change ((fmap (approx n) oo fmap (approx (S n))) wx = f2).
    rewrite fmap_comp.
    replace (approx n oo approx (S n)) with (approx n); auto.
    extensionality x.
    unfold compose.
    change (approx n (approx (S n) x)) with ((approx n oo approx (1 + n)) x).
    rewrite <- (approx_approx1 1 n).
    trivial.
    replace y' with (squash (n,(fmap (approx (S n)) wy))); auto.
    apply unsquash_inj.
    rewrite unsquash_squash, H1.
    apply injective_projections; simpl; auto.
    change ((fmap (approx n) oo fmap (approx (S n))) wy = f1).
    rewrite fmap_comp.
    replace (approx n oo approx (S n)) with (approx n); auto.
    rewrite H5.
    rewrite <- (unsquash_approx H1); auto.
    extensionality x.
    unfold compose.
    change (approx n (approx (S n) x)) with ((approx n oo approx (1 + n)) x).
    rewrite <- (approx_approx1 1 n).
    trivial.
  Qed.

  Lemma unage_identity : 
    forall x z, age x z -> identity z -> identity x.
  Proof.
    intros.
    rewrite identity_unit_equiv.
    rewrite identity_unit_equiv in H0.
    unfold unit_for in *.
    hnf in H; simpl in H.
    rewrite knot_age1 in H.
    case_eq (unsquash x); intros.
    rewrite H1 in H.
    destruct n; try discriminate.
    inversion H.
    subst z.
    destruct H0.
    repeat rewrite unsquash_squash in H0, H2.
    simpl in *.
    hnf; split; simpl.
    split; auto.
    rewrite H1.
    simpl.
    change (unit_for (H:=F_sa pred_sa) f f).
    rewrite <- (identity_unit_equiv f).
    apply F_identities with predicate pred_sa (approx n).
    rewrite identity_unit_equiv.
    apply H2.
  Qed.    

  Theorem knot_asa : @ASA knot knot_sa K.ageable_knot.
  Proof.
    constructor.
    exact unage_identity.
    exact age_join1.
    exact age_join2.
    exact unage_join1.
    exact unage_join2.
  Qed.

End KnotSa.
