(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import sepalg.
Require Import sepalg_generators.

Section unmaps.
  Variables A B:Type.
  Variable A_sa:sepalg A.
  Variable B_sa:sepalg B.

  Definition unmap_left (f:A -> B) :=
    forall x' y z,
      join x' (f y) (f z) ->
      { x:A & { y0:A | join x y0 z /\ f x = x' /\ f y0 = f y }}.

  Definition unmap_right (f:A -> B) :=
    forall x y z',
      join (f x) (f y) z' ->
      { y0: A & { z:A | join x y0 z /\ f y0 = f y /\ f z = z' }}.

End unmaps.
Implicit Arguments unmap_right.
Implicit Arguments unmap_left.

Section preserves_unmaps.
  Variable F : Type -> Type.
  Variable F_sa : forall A, sepalg A -> sepalg (F A).
  Variable fmap : forall A B, (A->B) -> F A -> F B.

  Record preserves_unmaps : Type :=
  { preserves_unmap_left :
       forall A B A_sa B_sa f
         (Hhom:join_hom f),
         unmap_left A_sa B_sa f ->
         unmap_left (F_sa A A_sa) (F_sa B B_sa) (fmap A B f)
  ; preserves_unmap_right :
       forall A B A_sa B_sa f
         (Hhom:join_hom f),
         unmap_right A_sa B_sa f ->
         unmap_right (F_sa A A_sa) (F_sa B B_sa) (fmap A B f)
  }.
End preserves_unmaps.

Section SepAlgProd_Functor.
  Variables X:Type.
  Variable Xsa : sepalg X. 

  Definition prod_left_fmap {A B} (f:A->B) (x: A*X) : B*X := (f (fst x), snd x).

  Lemma prod_left_fmap_id : forall A, prod_left_fmap (id A) = id (A*X).
  Proof.
    intros. extensionality x. destruct x; simpl. auto.
  Qed.

  Lemma prod_left_fmap_comp : forall A B C (f:A->B) (g:B->C),
    prod_left_fmap g oo prod_left_fmap f = prod_left_fmap (g oo f).
  Proof.
    intros. extensionality x. destruct x; simpl; auto.
  Qed.

  Lemma prod_preserves_unmaps : preserves_unmaps (fun A => prod A X) (fun A Asa => sa_prod Asa Xsa) (@prod_left_fmap).
  Proof.
    constructor; intros.
    
    hnf; intros.
    hnf in X0.
    destruct x' as [x'1 x'2].
    destruct y as [y1 y2].
    destruct z as [z1 z2].
    simpl in *.
    destruct H. simpl in *.
    apply X0 in H.
    destruct H as [x [y0 [? [? ?]]]].
    exists (x,x'2). exists (y0,y2).
    intuition; hnf; simpl; auto.
    unfold prod_left_fmap; simpl. congruence.
    unfold prod_left_fmap; simpl. congruence.

    hnf; intros.
    hnf in X0.
    destruct x as [x1 x2].
    destruct y as [y1 y2].
    destruct z' as [z'1 z'2].
    destruct H. simpl in*.
    apply  X0 in H.
    destruct H as [y0 [z [? [? ?]]]].
    exists (y0,y2). exists (z,z'2).
    intuition; hnf; simpl; auto.
    unfold prod_left_fmap; simpl. congruence.
    unfold prod_left_fmap; simpl. congruence.
  Qed.
End SepAlgProd_Functor.

Section SepAlgFun_Functor.
  Variable key: Type.

  Definition fun_fmap (A B:Type) (f:A->B) (x:key -> A) : key -> B := f oo x.
  Implicit Arguments fun_fmap.

  Lemma fun_fmap_id : forall A, fun_fmap (id A) = id (key -> A).
  Proof.
    unfold fun_fmap; intros.
    extensionality x y; auto.
  Qed.

  Lemma fun_fmap_comp : forall A B C (f:A->B) (g:B->C),
    fun_fmap g oo fun_fmap f  = fun_fmap (g oo f).
  Proof.
    unfold fun_fmap; intros.
    extensionality x y; auto.
  Qed.

  Lemma fun_preserves_unmaps : preserves_unmaps (fun A => key -> A) (sa_fun key) fun_fmap.
  Proof.
    constructor; intros.

    (* unmap left *)
    unfold fun_fmap, compose.
    red; intros.
    hnf in H.
    red in X.
    set (x := fun k => projT1 (X (x' k) (y k) (z k) (H k))).
    set (y0 := fun k => projT1 (projT2 (X (x' k) (y k) (z k) (H k)))).
    exists x; exists y0.
    split.
    hnf; intro k.
    unfold x, y0.
    destruct (X (x' k) (y k) (z k) (H k)) as [a [b [? [? ?]]]]; simpl; auto.
    split; extensionality k; unfold x, y0;
      destruct (X (x' k) (y k) (z k) (H k)) as [a [b [? [? ?]]]]; simpl; auto.
    
    (* unmap_right *)
    unfold fun_fmap, compose.
    red; intros.
    hnf in H.
    red in X.
    set (y0 := fun k => projT1 (X (x k) (y k) (z' k) (H k))).
    set (z  := fun k => projT1 (projT2 (X (x k) (y k) (z' k) (H k)))).
    exists y0.
    exists z.
    split.
    hnf; intro k.
    unfold y0, z;
      destruct (X (x k) (y k) (z' k) (H k)) as [a [b [? [? ?]]]]; simpl; auto.
    split.
    extensionality k; unfold y0, z;
      destruct (X (x k) (y k) (z' k) (H k)) as [a [b [? [? ?]]]]; simpl; auto.
    extensionality k; unfold y0, z;
      destruct (X (x k) (y k) (z' k) (H k)) as [a [b [? [? ?]]]]; simpl; auto.
  Qed.
End SepAlgFun_Functor.


Section SepAlgSigma_Functor.
  Variable I:Type.
  Variable SigmaF: I -> Type -> Type.
  Variable SigmaF_j : forall A, sepalg A -> forall (i:I), sepalg (SigmaF i A).
  Variable SigmaF_fmap : forall (i:I) A B, (A -> B) -> (SigmaF i A) -> (SigmaF i B).

  Let S A := { i:I & SigmaF i A }.
  Let injS A := existT (fun x => SigmaF x A).

  Hypothesis inj_sig_I : forall A i x y, injS A i x = injS A i y -> x = y.

  Hypothesis SigmaF_fmap_id : forall A (i:I),
    SigmaF_fmap i A A (id A) = id (SigmaF i A).
  Hypothesis SigmaF_fmap_comp : forall (i:I) A B C (f:A->B) (g:B->C),
    SigmaF_fmap i _ _ g oo SigmaF_fmap i _ _ f = SigmaF_fmap i _ _ (g oo f).

  Definition sigma_fmap A B (f:A->B) (x:S A) : S B :=
    match x with
    | existT i x' => injS B i (SigmaF_fmap i _ _ f x')
    end.

  Lemma sigma_fmap_id : forall A,
    sigma_fmap A A (id A) = id (S A).
  Proof.
    unfold sigma_fmap; intros; extensionality x.
    destruct x; unfold injS; simpl.
    rewrite SigmaF_fmap_id.
    auto.
  Qed.

  Lemma sigma_fmap_comp : forall A B C (f:A->B) (g:B->C),
    sigma_fmap _ _ g oo sigma_fmap _ _ f = sigma_fmap _ _ (g oo f).
  Proof.
    unfold sigma_fmap; intros; extensionality x.
    destruct x; unfold injS; simpl.
    rewrite <- SigmaF_fmap_comp.
    auto.
  Qed.

  Lemma sigma_preserves_unmaps :
    (forall i, preserves_unmaps (SigmaF i) (fun A A_sa => SigmaF_j A A_sa i) (SigmaF_fmap i)) ->
    preserves_unmaps S 
      (fun (A:Type) A_sa => (sa_sigma (fun i => SigmaF i A) (SigmaF_j A A_sa)) (inj_sig_I A)) sigma_fmap.
  Proof.
    intros.
    constructor; intros.

    red; intros.
    destruct z as [i z]; simpl in *.
    destruct x' as [xi x']; simpl in *.
    destruct y as [yi y]; simpl in *.
    assert (i = xi /\ xi = yi).
    inv H; simpl; auto.
    destruct H0; subst xi yi.
    destruct (X i) as [H1 H2].
    destruct (H1 _ _ _ _ _ Hhom X0 x' y z) as [u [v [? [? ?]]]].
    inv H; auto.
    generalize (inj_sig_I _ _ _ _ H4); clear H4.
    generalize (inj_sig_I _ _ _ _ H5); clear H5.
    generalize (inj_sig_I _ _ _ _ H6); clear H6.
    intros; subst; auto.
    exists (injS A i u).
    exists (injS A i v).
    subst; simpl; intuition.
    constructor; auto.
    rewrite H4; auto.

    red; intros.
    destruct x as [i x].
    destruct y as [yi y].
    destruct z' as [zi z'].
    assert (i = yi /\ yi = zi).
    inv H; auto.
    destruct H0; subst zi yi.
    destruct (X i) as [H1 H2].
    destruct (H2 _ _ _ _ _ Hhom X0 x y z') as [y0 [z [? [? ?]]]].
    inv H; auto.
    generalize (inj_sig_I _ _ _ _ H4); clear H4.
    generalize (inj_sig_I _ _ _ _ H5); clear H5.
    generalize (inj_sig_I _ _ _ _ H6); clear H6.
    intros; subst; auto.
    exists (injS A i y0).
    exists (injS A i z).
    subst; simpl; intuition.
    constructor; auto.
    rewrite H3; auto.
  Qed.
End SepAlgSigma_Functor.

Section SepAlgProp_Functor.
  Variable F:Type -> Type.
  Variable Fj : forall A, sepalg A -> sepalg (F A).
  Variable fmap : forall A B, (A -> B) -> F A -> F B.

  Variable fmap_id : forall A, fmap _ _ (id A) = id (F A).
  Variable fmap_comp : forall A B C (f:A->B) (g:B->C),
    fmap _ _ g oo fmap _ _ f = fmap _ _ (g oo f).

  Variable P : forall A, F A -> Prop.

  Variable HPunit :
    forall A (A_sa: sepalg A) x e, P A x -> @unit_for _ (Fj A A_sa) e x -> P A e.

  Variable HPjoin : forall A A_sa x y z,
    @join _ (Fj A A_sa) x y z -> P A x -> P A y -> P A z.

  Variable HPfmap1 : forall A B f x, 
    P A x -> P B (fmap A B f x).
  Variable HPfmap2 : forall A B f x, 
    P B (fmap A B f x) -> P A x.

  Let S A := { x:F A | P A x }.
  
  Definition prop_fmap (A B:Type) (f:A->B) (x:S A) : S B :=
    match x with
    | exist x' H => exist (fun x => P B x) (fmap A B f x') (HPfmap1 A B f x' H)
    end.

  Lemma prop_fmap_id : forall A, prop_fmap A A (id A) = id (S A).
  Proof.
    unfold prop_fmap; intros; extensionality x.
    destruct x; simpl; auto.
    generalize (HPfmap1 A A (id A) x p).
    rewrite fmap_id; intro.
    replace p0 with p by (apply proof_irr); auto.
  Qed.

  Lemma prop_fmap_comp : forall A B C (f:A->B) (g:B->C),
    prop_fmap _ _ g oo prop_fmap _ _ f = prop_fmap _ _ (g oo f).
  Proof.
    unfold prop_fmap; intros; extensionality x.
    unfold compose.
    destruct x; simpl.
    generalize (HPfmap1 _ _ g (fmap _ _ f x) (HPfmap1 _ _ f x p)).
    change (fmap _ _ g (fmap _ _ f x))
      with ((fmap _ _ g oo fmap _ _ f) x).
    rewrite fmap_comp.
    intros.
    replace (HPfmap1 _ _ (fun x => g (f x)) x p) with p0 by (apply proof_irr).
    auto.
  Qed.

  Hypothesis F_preserves_unmaps : preserves_unmaps F Fj fmap.

  Lemma prop_preserves_unmaps :
      preserves_unmaps S 
     (fun A A_sa => sa_prop (Fj A A_sa) (P A) (HPunit A A_sa) (HPjoin A A_sa)) prop_fmap.
  Proof.
    destruct F_preserves_unmaps as [H1 H2].
    constructor; intros.

    red; intros.
    destruct x' as [x' Hx'].
    destruct y as [y Hy].
    destruct z as [z Hz].
    simpl in *.
    destruct (H1 _ _ _ _ _ Hhom X x' y z) as [x [y0 [? [? ?]]]]; auto.
    subst x'.
    exists (existT (fun x => P A x) x (HPfmap2 _ _ f x Hx')).
    assert (P A y0).
    eapply HPfmap2 with B f.
    rewrite H4.
    apply HPfmap1; auto.
    exists (existT (fun x => P A x) y0 H3).
    simpl; intuition.
    replace (HPfmap1 _ _ f x (HPfmap2 _ _ f x Hx')) with Hx'
      by (apply proof_irr); auto.
    generalize (HPfmap1 A B f y0 H3).
    rewrite H4; intros.
    replace (HPfmap1 A B f y Hy) with p
      by (apply proof_irr); auto.
    
    red; intros.
    destruct x as [x Hx].
    destruct y as [y Hy].
    destruct z' as [z' Hz'].
    simpl in *.
    destruct (H2 _ _ _ _ _ Hhom X x y z') as [y0 [z [? [? ?]]]]; auto.
    subst z'.
    assert (P A y0).
    apply HPfmap2 with B f.
    rewrite H3.
    apply HPfmap1; auto.
    exists (existT (fun x => P A x) y0 H4).
    exists (existT (fun x => P A x) z (HPfmap2 _ _ f z Hz')).
    simpl; intuition.
    generalize (HPfmap1 A B f y0 H4).
    rewrite H3.
    intro; replace (HPfmap1 A B f y Hy) with p
      by (apply proof_irr); auto.
    replace (HPfmap1 _ _ f z (HPfmap2 _ _ f z Hz')) with Hz'
      by (apply proof_irr); auto.
  Qed.
End SepAlgProp_Functor.

Section SepAlgList_Functor.

  Lemma list_fmap_id : forall A, map (id A) = id (list A).
  Proof.
    intro A; extensionality l; induction l; simpl; auto.
    rewrite IHl; auto.
  Qed.

  Lemma list_fmap_comp : forall A B C (f:A->B) (g:B->C),
    map g oo map f = map (g oo f).
  Proof.
    intros; extensionality l; induction l; simpl; auto.
    unfold compose in *; simpl.
    rewrite IHl; auto.
  Qed.

  Lemma list_preserves_unmaps :
    preserves_unmaps list list_sa map.
  Proof.
    constructor; intros.

    red; intro x'; induction x'; simpl; intros.
    exists nil; exists nil.
    inv H.
    destruct z; try discriminate.
    simpl; intuition.
    destruct y; try discriminate.
    constructor.

    destruct y; destruct z;
      try (elimtype False; inversion H; fail).
    destruct (IHx' y z) as [x [y0 [? [? ?]]]].
    inv H; auto.
    destruct (X a a0 a1) as [u [v [? [? ?]]]].
    inv H; auto.
    exists (u::x); exists (v::y0); simpl; intuition.
    inv H.
    constructor; auto.
    congruence.
    congruence.

    red; intro x'; induction x'; simpl; intros.
    exists nil; exists nil.
    inv H.
    simpl; intuition; constructor.
    destruct y; destruct z';
      try (elimtype False; inv H; fail).
    destruct (IHx' y z') as [y0 [z [? [? ?]]]].
    inv H; auto.
    destruct (X a a0 b) as [u [v [? [? ?]]]].
    inv H; auto.
    exists (u::y0); exists (v::z); simpl; intuition.
    constructor; auto.
    congruence.
    congruence.
  Qed.
End SepAlgList_Functor.
